/* COPYRIGHT (C) GUPTA TECHNOLOGIES, INC. 1984-1990 */
/*
  Copyright (c) 1989 Gupta Technologies, Inc.
  All Rights Reserved.

Program Name:
  TESTWIN

Function Description:
  The is a sample MSWindows program to demonstrate some
  of the interfaces to SQLBase.

Revision History:
  04/04/86 DAI original
  07/12/87 DBS change SQLDES interface for column header pointer
  08/15/87 DBS packed decimal
  04/22/88 DBS change SQLCON to SQLCNC
  01/26/89 JMT Reorganized and redocumented code.
	       Used an EDIT control to display logged text.
  05/04/90 DBS bug fix: add default window sizes
*/


/* Include Files */

#define LINT_ARGS
#define NOMINMAX

#include "sql.h"
#include <stdio.h>
#include <memory.h>
#include <windows.h>
#include <io.h>
#include <stdlib.h>
#include <string.h>


/* Constants */

#define TESTCLASSNAME "SQL:TEST"
#define TESTWINDOWCAPTION "Example SQL interface"
#define TESTINITMSG  "Its likely the database was not initialized\n"
#define TESTCOPYMSG  \
"Its likely the database was not initialized.  To initialize, copy START.DBS to \\SQLBASE\\DEMO\\DEMO.DBS"

/* Global Variables */

extern	  int	    __argc;
extern	  char**    __argv;

HANDLE		hInst;
HANDLE		hgWndLog;
BOOL		bgOutOfSpace = FALSE;
BOOL		bgQueryOnly = FALSE;
BOOL		bgSQLInit = FALSE;
int		gNumRuns = 1;		/* Number of runs		    */
int		bgExitFlag = FALSE;
int		bgExitAllFlag = FALSE;
char*		szgLogFileName = "TESTWIN.LOG";
char*		szgDBName      = "DEMO";
SQLTCUR 	gCursor = 0;		/* SQLBase cursor		    */
SQLTRCD 	gRetCode = 0;		/* error number 		    */

char *szgCreateTableCmd =
	"CREATE TABLE X (A NUMBER, B DATETIME, C CHAR(30), D NUMBER)";
char *gszInsertCmd =
	"INSERT INTO X(A, B, C, D) VALUES (:1, :2, :3, :4)";
char *szgCreateIndexCmd =
	"CREATE INDEX IX_B ON X(B)";
char *gszSelectCmd =
	"SELECT B, A, D, C FROM X ORDER BY 1";


/* Function Prototypes */

#ifdef WIN21  /* set this to 1 if running in Windows 2.1		    */

BOOL  FAR PASCAL ExitWindows(DWORD dwReserved, WORD wReturnCode);

#endif

int PASCAL WinMain(HANDLE,HANDLE,LPSTR,int);
BOOL TestInit(VOID);
long FAR PASCAL TestWndProc(HWND,unsigned,WORD,LONG);
VOID TestLogMessage(char*,...);
BOOL TestSQLSample(HWND);
int FAR PASCAL TestYieldProc(VOID);
VOID TestLogSQLFailure(char *);
VOID TestParseCmdLine(VOID);
BOOL TestSQLSelect(VOID);
BOOL TestSQLInsert(VOID);


/*

Function Name:
  WinMain

Parameters:
  HANDLE    Identifies this instance of TESTWIN.EXE.
  HANDLE    Identifies the previous instance (It may be NULL).
  LPSTR     Pointer to the contents of the command line.
  int	    Describes how the window should be shown--iconized,
	    overlapped, etc.
Returns:
  int	    Application exit code.

Description:
  This is the main entry point to TESTWIN.EXE.

*/

int PASCAL WinMain(hInstance, hPrevInstance, lpszCmdLine, cmdShow)
HANDLE hInstance;
HANDLE hPrevInstance;
LPSTR  lpszCmdLine;
int	    cmdShow;
{
	  MSG	    msg;
	  HWND	    hWnd;

  hInst = hInstance;

  if( ! hPrevInstance  &&  ! TestInit() )
    return FALSE;

  hWnd = CreateWindow(TESTCLASSNAME,
		      TESTWINDOWCAPTION,
		      WS_TILEDWINDOW,
		      0,		/*  x - ignored for tiled windows   */
		      0,		/*  y - ignored for tiled windows   */
		      CW_USEDEFAULT,	/* cx - ignored for tiled windows   */
		      CW_USEDEFAULT,	/* cy - ignored for tiled windows   */
		      NULL,		/* no parent			    */
		      NULL,		/* use class menu		    */
		      hInst,	/* handle to window instance		    */
		      NULL);		/* no params to pass on 	    */


  ShowWindow(hWnd,cmdShow);
  UpdateWindow(hWnd);

  /*  process command line options */
  TestParseCmdLine();

  /* call SQLBase API */
  TestSQLSample(hWnd);

  /* close log file */
  TestLogMessage(NULL);

  if( bgExitFlag )
    /* exit this program */
    PostQuitMessage( 0 );

  /* Poll Messages From Queue */

  while( GetMessage( &msg, NULL, 0, 0 ) )
  {
    TranslateMessage( &msg );
    DispatchMessage( &msg );
  }
  return msg.wParam;

}


/*

Function Name:
  TestInit

Parameters:
  None.

Returns:
  BOOL	   TRUE on success, FALSE on failure.

Description:
  Initializes the first instance of TESTWIN.EXE by registering
  the class of the main window.

*/

BOOL  TestInit(VOID)
{
  WNDCLASS  Class;

  Class.style	      = CS_HREDRAW | CS_VREDRAW;
  Class.lpfnWndProc   = (long (FAR PASCAL *)())TestWndProc;
  Class.cbClsExtra    = 0;
  Class.cbWndExtra    = 0;
  Class.hInstance     = hInst;
  Class.hIcon	      = NULL;
  Class.hCursor       = LoadCursor(NULL,IDC_ARROW);
  Class.hbrBackground = COLOR_WINDOW + 1;
  Class.lpszMenuName  = (LPSTR) NULL;
  Class.lpszClassName = (LPSTR)TESTCLASSNAME;
  return RegisterClass( &Class );
}

/*

Function Name:
  TestWndProc

Parameters:
  HWND		Handle of the main window.
  unsigned	Identifies the message.
  WORD		More message information.
  LONG		More message information.

Returns:
  LONG		Return value depends on the message received.

Description:
  This is the window procedure for the main window.

*/

long FAR PASCAL TestWndProc( hWnd, message, wParam, lParam )
HWND	    hWnd;
unsigned    message;
WORD	    wParam;
LONG	    lParam;
{

  switch( message )
  {
    case WM_CREATE:
      /* Create a child EDIT window that will be used to log messages
	 to the screen.
  */
      hgWndLog = CreateWindow("EDIT",
		      "",
		      WS_CHILD | WS_VISIBLE | WS_VSCROLL | WS_HSCROLL |
		      ES_LEFT | ES_MULTILINE |
		      ES_AUTOHSCROLL | ES_AUTOVSCROLL,
		      0,
		      0,
		      0,
		      0,
		      hWnd,
		      1,
		      hInst,
		      NULL);
      break;

    case WM_DESTROY:
      if (bgSQLInit)
	/* free resources allocated to dynamic API library		    */
	sqldon();
      PostQuitMessage( 0 );
      break;

    case WM_SIZE:
      /* When the main window is sized, adjust the size of the EDIT window
	 so that it occupies the entire client area.
      */
      MoveWindow (hgWndLog, 0, 0, LOWORD(lParam), HIWORD(lParam), TRUE);
      break;

    case WM_COMMAND:
      if (wParam == 1 && HIWORD (lParam) == EN_ERRSPACE)
      {
	/* There is not enough memory to store additional text in the
	   EDIT window.
	*/
	MessageBox(hWnd, "Edit control out of space.",
		      TESTWINDOWCAPTION, MB_OK | MB_ICONHAND);
	/* set a flag so that no more text is inserted			    */
	bgOutOfSpace = TRUE;
      }
      break;

    default:
      return DefWindowProc( hWnd, message, wParam, lParam );
      break;
  }

  return(0L);
}


/*

Function Name:
  TestParseCmdLine

Parameters:
  VOID

Returns:
  VOID

Description:
  Searches command line arguments for switches that control
  the behavior of this program.  Modifies global variables based on
  the content of switches.

*/

VOID TestParseCmdLine(VOID)
{
  char*     p;

  /* adjust argument list  */
  __argc--;
  __argv++;
  /* scan argument list  */
  while (__argc-- > 0)
  {
    /* fetch argument */
    p = *__argv++;
    /* option? */
    if (*p == '-')
    {
      /* scan selected options */
      while (*++p) switch (*p)
      {
	/* query-only option */
	case 'q':
	  bgQueryOnly = TRUE;
	break;
	/* number of runs option */
	case 'r':
	if (p[1] >= '0' && p[1] <= '9')
	  {
	    gNumRuns = atoi(p + 1);
	    p+=strlen(&p[1]);  /* skip rest of arg			    */
	  }
	else if (__argc-- > 0)
	    gNumRuns = atoi(*__argv++);
	else
	    TestLogMessage("missing argument\n");
	  if (gNumRuns < 1)
	    gNumRuns = 1;
	break;
	/* output file option */
      case 'o':
	if (__argc-- > 0)
	    szgLogFileName = *__argv++;
	else
	    TestLogMessage("missing argument\n");
	break;
	/* exit windows option */
      case 'E':
	  bgExitAllFlag = TRUE;
	/* fall through */
	/* exit program option */
      case 'e':
	  bgExitFlag = TRUE;
	break;
	default:
	  TestLogMessage("Invalid option -%c\n", *p);
	break;
      }
    }
  }
}



/*

Function Name:
  TestSQLSample

Parameters:
  HWND	   Handle of the main window

Returns:
  BOOL	   TRUE for success, FALSE for failure.

Description:
  This procedure performs the following steps (command line options can
  be used to skip steps or modify the behavior of steps):

    The database is connected to by calling  sqlcon  and
    passing the database name string.

    The table is created  by  calling	sqlcex
    with the CREATE TABLE string.

    TestSQLInsert is called to insert fields read from a file into the table.

    The index is created  by calling sqlcex with the CREATE INDEX string.

    TestSQLSelect is called to select the fields inserted into the table.

    The database is disconnected from by calling sqldis.

    dynamic librarys resources are freed by calling sqldon
*/

BOOL	  TestSQLSample(hWnd)
HWND	    hWnd;
{
int	  nRuns;	    /* number of runs				    */
FARPROC   lpProcInst;
char	  cMsgBuffer[200];
BOOL	  bStatus;	    /* return code				    */

  bStatus = FALSE;
  lpProcInst = NULL;

  /* initialize cursor number	    */;
  gCursor = 0;

  /* initialize dynamic API library */
  if (gRetCode =
      sqlini(lpProcInst = MakeProcInstance(TestYieldProc, hInst)))
  {
    TestLogMessage("Cannot initialize API interface - %u\n", gRetCode);
    goto errexit;
  }
  else
    bgSQLInit = TRUE;

  TestLogMessage("Connecting to database\n");

  /* set number of pages */
  sqlscp(15);

  /* connect To the database */
  if (gRetCode = sqlcnc(&gCursor, szgDBName, 0))
  {
    TestLogSQLFailure("CONNECT");
    TestLogMessage(TESTINITMSG);
    MessageBox(hWnd, TESTCOPYMSG, TESTWINDOWCAPTION, MB_OK | MB_ICONHAND);
    goto errexit;
  }

  /* not doing query only?	    */
  if (!bgQueryOnly)
  {
    TestLogMessage("%s\n", szgCreateTableCmd);

    /* Compile and execute create table */
    if (sqlcex(gCursor, szgCreateTableCmd, 0))
  {
      TestLogSQLFailure("CREATE TABLE");
      TestLogMessage(TESTINITMSG);
      MessageBox(hWnd, TESTCOPYMSG, TESTWINDOWCAPTION, MB_OK | MB_ICONHAND);
      goto errexit;
    }


    /* Insert records from input file */
    if (!TestSQLInsert())
      goto errexit;


    TestLogMessage("%s\n", szgCreateIndexCmd);

    /* create the index */
    if (sqlcex(gCursor, szgCreateIndexCmd, 0))
  {
      TestLogSQLFailure("CREATE INDEX");
      goto errexit;
    }

    TestLogMessage("COMMIT\n");

    /* commit the changes */
    if (sqlcmt(gCursor))
    {
      TestLogSQLFailure("COMMIT");
      goto errexit;
    }
  }  /* end of if (!bgQueryOnly) */

  /* select fields from table */
  for (nRuns = 0; nRuns < gNumRuns; nRuns++)
    if (!TestSQLSelect())
      goto errexit;


  TestLogMessage("Disconnect from database\n");
  /* disconnect  from the database */
  if (gRetCode = sqldis(gCursor))
  {
    /* indicate invalid cursor	    */
    gCursor = 0;
    TestLogSQLFailure("DISCONNECT");
    goto errexit;
  }

  bStatus = TRUE;

errexit:

  if (bgSQLInit)
      {
    /* free resources allocated to dynamic API library			    */
    sqldon();
    bgSQLInit = FALSE;
      }

  if (lpProcInst)
    FreeProcInstance(lpProcInst);

  sprintf(cMsgBuffer, "Output has also been printed to file \"%s\"\n",
	  szgLogFileName);
  TestLogMessage(cMsgBuffer);
  TestLogMessage("*****\n");

  return bStatus;
}


/*

Function Name:
  TestSQLInsert

Parameters:
  VOID

Returns:
  BOOL	   TRUE for success, FALSE for failure.

Description:
  This procedure performs the following steps:

    The INSERT command	is  compiled.

    The  DATA file is opened.

    Lines are read from the  DATA  file.

    For  each  line  the commas in the line are located to separate
    each  field  value.   Sqlbnn  is called to bind each field value
    to the appropriate bind variable.  After the input line has been
    scanned and the data has been bound, sqlexe is called to execute
    the INSERT command to insert the row/data into the	database.   This
    is done until the end of the DATA file is reached.

    The  DATA file is closed.

*/

BOOL TestSQLInsert(VOID)
{
FILE*	  fpIn; 		/* input file				    */
int	  nRows;		/* number of rows			    */
SQLTNBV   nBindVar;		/* number of bind variables		    */
char	  cLineBuffer[200];	/* I/O line				    */
LPSTR	  lpChar;		/* character pointer			    */
LPSTR	  lpScan;		/* scan pointer 			    */
SQLTCDL   nLength;		/* length				    */
SQLTSLC   nCol; 		/* column number			    */
BOOL	  bStatus;		/* return code				    */

  fpIn = 0;
  bStatus = FALSE;

  TestLogMessage("%s\n", gszInsertCmd);
  /* compile the insert command */
  if (sqlcom(gCursor, gszInsertCmd, 0))
    {
    TestLogSQLFailure("INSERT COMPILE");
    goto errexit;
    }

  /* open the input file */
  if (!(fpIn = fopen("DATA", "r")))
    {
    TestLogMessage("Cannot open data file\n");
    goto errexit;
    }

  /* initialize number of rows	    */
  nRows = 0;

  /* get number of bind variables     */
  sqlnbv(gCursor, &nBindVar);

  /* scan the input file	    */
  while (fgets(cLineBuffer, sizeof(cLineBuffer), fpIn))
    {
    /* allow other windows programs to run				    */
    TestYieldProc();

    /* increment number of rows     */
    nRows++;

    /* remove new line		    */
    cLineBuffer[strlen(cLineBuffer) - 1] = 0;

    /* scan the line		    */
    for (lpChar = (LPSTR)cLineBuffer, nCol = 1;nCol <= nBindVar; nCol++)
      {
      /* scan string for comma	    */
      for (lpScan = lpChar;*lpScan && *lpScan != ','; lpScan++);
      /* compute length 	    */
      nLength = (SQLTCDL)(lpScan - lpChar);
      /* bind by number 	    */
      if (sqlbnn(gCursor, nCol, lpChar, nLength, 0, SQLPBUF))
	{
	TestLogSQLFailure("BIND");
	goto errexit;
	}
      /* locate end		    */
      lpChar += nLength;
      /* look for comma 	    */
      if (*lpChar == ',')
	lpChar++;
      }

    /* execute	insert */
    if (sqlexe(gCursor))
      {
      TestLogMessage("%d ROWS INSERTED\n",nRows);
      TestLogSQLFailure("INSERT EXECUTE");
      goto errexit;
      }
    }

  bStatus = TRUE;

errexit:

  if (fpIn)
    fclose( fpIn );

  return bStatus;
}


/*

Function Name:
  TestSQLSelect

Parameters:
  VOID

Returns:
  BOOL	   TRUE for success, FALSE for failure.

Description:
  This procedure performs the following steps:

    The SELECT command is compiled by calling sqlcom.

    For each column in the SELECT command SQLDES  is called
    to get descriptive information.

    Once the descriptive information is gotten, sqlssb is called to set
    the buffer of where the select data will be returned.

    The SELECT command is  executed  by  calling  sqlexe.
    Rows  are fetched from the SELECT command until the end of the
    SELECT  is	reached.  This is done by repeatedly calling sqlfet.
    Each  row  is  displayed.
*/

BOOL TestSQLSelect(VOID)
{
LPSTR	  lpChar;		/* global memory  pointer		    */
SQLTCDL   nLength;		/* field length 			    */
char	  cLocalBuffer[200];	/* Local memory I/O line		    */
PSTR	  pChar;		/* local memory pointer 		    */
SQLTSLC   nCol; 		/* column number			    */
SQLTPDL   nProgDataLen; 	/* program buffer length		    */
SQLTDDT   nDataType;		/* database data type			    */
SQLTNSI   nSelItems;		/* number of select items		    */
BOOL	  bStatus;		/* return code				    */
int	  nChar;		/* Index to characters			    */
HANDLE	  hSelectBuffer;	/* handle to fetch buffer memory	    */
LPSTR	  lpSelectBuffer;	/* pointer to fetch buffer memory	    */

  bStatus = FALSE;
  hSelectBuffer = NULL;
  lpSelectBuffer = NULL;

  TestLogMessage("%s\n", gszSelectCmd);
  /* compile the select command */
  if (sqlcom(gCursor, gszSelectCmd, 0))
    {
    TestLogSQLFailure("SELECT COMPILE");
    goto errexit;
    }

  /* allocate select buffer in global memory				    */
  hSelectBuffer = GlobalAlloc(GHND, (LONG)200);
  if (!hSelectBuffer)
    goto errexit;
  /* lock global memory */
  lpSelectBuffer = GlobalLock(hSelectBuffer);
  if (!lpSelectBuffer)
    goto errexit;

    /*
  Get descriptive information about select.
    */

  /* set pointer to fetch buffer    */
  lpChar = lpSelectBuffer;

  /* get number of select items     */
  sqlnsi(gCursor, &nSelItems);

  /* get information on each column   */
  for (nCol = 1; nCol <= nSelItems; nCol++)
    {
    if (gRetCode = sqldes(gCursor, nCol, &nDataType,
	  &nProgDataLen, SQLNPTR, SQLNPTR, SQLNPTR,
	  SQLNPTR))
      {
      TestLogSQLFailure("SELECT DESCRIBE");
      goto errexit;
      }

    switch (nDataType)
      {
	case SQLDNUM:
	nProgDataLen = 7;
	  break;

	case SQLDDAT:
	/* set length to date-time size */
	nProgDataLen = SQLSCDA;
	  break;
      }

    /* set the select buffer */
    if (gRetCode =
	  sqlssb(gCursor, nCol, SQLPBUF, lpChar, nProgDataLen,
			0, SQLNPTR, SQLNPTR))
      {
      TestLogSQLFailure("SET SELECT BUFFER");
      goto errexit;
      }
    /* locate next area 	    */
    lpChar += (nProgDataLen + 1);
    }

  /* execute the select command */
  if (sqlexe(gCursor))
    {
    TestLogSQLFailure("SELECT EXECUTE");
    goto errexit;
    }

    /*
    Fetch and display the data.
    */

  /* compute the length 	    */
  nLength = (SQLTCDL)(lpChar - lpSelectBuffer);
  /* concatenate a zero to the string */
  *lpChar = 0;

    for (;;)
    {
    /* fill the line with spaces    */
    for (nChar=0; nChar<nLength; nChar++)
      lpSelectBuffer[nChar] = ' ';
    /* perform fetch */
    if (gRetCode = sqlfet(gCursor))
      /* failure or end of file     */
	break;
    /* store the line in local memory */
    lpChar = lpSelectBuffer;
    pChar  = cLocalBuffer;
    while (*pChar++ = *lpChar++);
    /* print the line		    */
    TestLogMessage("%s\n",cLocalBuffer);
    }

  /* test for failure on fetch						    */
  if (gRetCode != 1)
    {
    TestLogSQLFailure("FETCH");
    goto errexit;
    }

  bStatus = TRUE;

errexit:
  if (lpSelectBuffer)
    GlobalUnlock(hSelectBuffer);

  if (hSelectBuffer)
    GlobalFree(hSelectBuffer);

  return bStatus;
}


/*

Function Name:
  TestYieldProc

Parameters:
  VOID

Returns:
  int	always 0

Description:
  TestYieldProc is called periodically by SQLBase when the multi-user
  server is processing an SQLBase request.  It gives other applications
  a chance to run.

  NOTE:  Call back procedures such as TestYieldProc
	 MUST be declared in the .DEF file.
*/

int FAR PASCAL TestYieldProc()
{
MSG	    msg;

  while( PeekMessage( &msg, NULL, 0, 0, TRUE ) )
  {
    TranslateMessage( &msg );
    DispatchMessage( &msg );
  }
  /* return successfully  */
  return 0;
}

/*

Function Name:
  TestLogMessage

Parameters:
  VOID

Returns:
  VOID

Description:
  Formats a message then writes it to a log file and displays it on the
  screen.  This procedure also opens and closes the log file.
*/

VOID	  TestLogMessage(szFormat,arg1,arg2,arg3)
char*	  szFormat;
char*	  arg1;
char*	  arg2;
char*	  arg3;
{
char*	    p1;
char*	    p2;
char	    cBuffer1[128];
char	    cBuffer2[128];
int	    hfile;

static	  FILE*     fpLog = NULL;

  if (fpLog == NULL)
    fpLog = fopen(szgLogFileName, "w");

  if (!szFormat)
  {
    if (fpLog)
    {
      /* flush output stream						    */
      fflush(fpLog);

      /* duplicate file descriptor */
      if ((hfile = dup(fileno(fpLog))) >= 0)
	/* force directory to be updated				    */
	close(hfile);
  }
    /* print no further 	    */
    return;
  }

  if (fpLog)
    fprintf(fpLog,szFormat,arg1,arg2,arg3);

  /* EDIT window out of space */
  if (bgOutOfSpace)
    return;

  /* szFormat string to buffer */
  sprintf(cBuffer1,szFormat,arg1,arg2,arg3);

  /* translate newlines to return,newline so that the
     string will be properly formatted by the EDIT window		    */

  p1 = cBuffer1;
  p2 = cBuffer2;
  do
  {
    if (*p1 == '\n')
      *p2++ = '\r';
    *p2++ = *p1;
  }
  while (*p1++);

  /* Append the string to the text in the EDIT control. ( This call assumes
     that the text cursor in the EDIT window follows the last character)    */

  SendMessage (hgWndLog, EM_REPLACESEL, 0, (LONG)(LPSTR)cBuffer2);
}


/*

Function Name:
  TestLogSQLFailure

Parameters:
  char *   Pointer to a null terminated string describing
	   the action that caused a failure.

Returns:
  VOID

Description:
  Displays information about an SQLBase failure.

*/

VOID	  TestLogSQLFailure(szAction)
char *szAction;
{
SQLTEPO   ErrorPos;		/* error position			    */

  if (gCursor)			/* cursor connected?			    */
  {
    sqlrcd(gCursor, &gRetCode); /* get the error			    */
    sqlepo(gCursor, &ErrorPos); /* get error position			    */
  }

  TestLogMessage("TESTWIN: failure on %s gRetCode=%d, ErrorPos=%d\n",
		szAction, gRetCode, ErrorPos);
}
